'use strict';

exports.__esModule = true;
exports.buildFieldId = buildFieldId;
exports.parseFieldId = parseFieldId;
exports.buildFieldIdFromUserKeyPath = buildFieldIdFromUserKeyPath;
exports.buildFieldDomId = buildFieldDomId;
exports.buildFieldDomIdWithValue = buildFieldDomIdWithValue;
exports.buildFieldPropsStateId = buildFieldPropsStateId;
exports.buildFieldValidationStateId = buildFieldValidationStateId;
exports.parseFieldStateId = parseFieldStateId;

var _Constant = require('../../enums/Constant');

var _Constant2 = _interopRequireDefault(_Constant);

var _buildErrorMessage = require('../buildErrorMessage');

var _buildErrorMessage2 = _interopRequireDefault(_buildErrorMessage);

var _utils = require('../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @desc Builds internal identificator of the field,
 *       e.g. `user.email`
 *
 */
function buildFieldId(keyPath) {
  return keyPath.join(_Constant2.default.FIELD_ID_DELIMITER);
}

/**
 * @desc Parses internal identificator of the field,
 *       and returns name of the attribute and key path to its value.
 *       e.g. `email` & `['user', 'email']`
 *
 */
function parseFieldId(fieldId) {
  var keyPath = fieldId.split(_Constant2.default.FIELD_ID_DELIMITER);
  return { keyPath: keyPath, attr: keyPath[keyPath.length - 1] };
}

/**
 * @desc Builds internal identificator of the field from a key path.
 *
 */
function buildFieldIdFromUserKeyPath(userKeyPath) {
  var normalizedKeyPath = (0, _utils.normalizeKeyPath)(userKeyPath);
  return buildFieldId(normalizedKeyPath);
}

/**
 * @desc Builds DOM id attribute of the field,
 *       including layerId to namespace it,
 *       e.g. `loginForm___user___email`
 *
 */
function buildFieldDomId(layerId, keyPath) {
  return [layerId].concat(keyPath).join(_Constant2.default.FIELD_DOM_ID_DELIMITER);
}

/**
 * @desc Builds DOM id attribute of the field,
 *       including layerId to namespace it
 *       and value in case if it's radio button,
 *       e.g. `signupForm___provider___facebook`
 *
 */
function buildFieldDomIdWithValue(layerId, userKeyPath, value) {
  if (!(0, _utils.isString)(value) && !(0, _utils.isNumber)(value)) {
    throw new Error((0, _buildErrorMessage2.default)({
      layerId: layerId,
      message: ['Value for dom id must be a string or a number.', value ? 'Passed value is: ' + value : null, 'Field key path: ' + JSON.stringify(userKeyPath)]
    }));
  }

  var normalizedKeyPath = (0, _utils.normalizeKeyPath)(userKeyPath);

  return [layerId].concat(normalizedKeyPath).concat(value).join(_Constant2.default.FIELD_DOM_ID_DELIMITER);
}

/**
 * @desc Builds internal identificator of the slice of the state,
 *       that holds DOM props for the field,
 *       e.g. `fieldPropsState---user.email`
 *
 */
function buildFieldPropsStateId(fieldId) {
  var prefix = _Constant2.default.FIELD_PROPS_STATE_ID_PREFIX;
  var _ = _Constant2.default.FIELD_STATE_ID_DELIMITER;

  return '' + prefix + _ + fieldId;
}

/**
 * @desc Builds internal identificator of the slice of the state,
 *       that holds validation results for the field,
 *       e.g. `fieldValidationState---user.email`
 *
 */
function buildFieldValidationStateId(fieldId) {
  var prefix = _Constant2.default.FIELD_VALIDATION_STATE_ID_PREFIX;
  var _ = _Constant2.default.FIELD_STATE_ID_DELIMITER;

  return '' + prefix + _ + fieldId;
}

/**
 * @desc Parses internal identificator of the slice of the state,
 *       and returns type of data & fieldId,
 *       e.g. `fieldValidationState` & `user.email`
 *
 */
function parseFieldStateId(fieldStateId) {
  var _fieldStateId$split = fieldStateId.split(_Constant2.default.FIELD_STATE_ID_DELIMITER),
      dataType = _fieldStateId$split[0],
      fieldId = _fieldStateId$split[1];

  return { dataType: dataType, fieldId: fieldId };
}