'use strict';

exports.__esModule = true;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.performFieldValidation = performFieldValidation;
exports.performOnMountValidation = performOnMountValidation;
exports.performOnSubmitValidation = performOnSubmitValidation;

var _lodash = require('../utils/lodash');

var _ = _interopRequireWildcard(_lodash);

var _Strategy = require('../../enums/Strategy');

var _Strategy2 = _interopRequireDefault(_Strategy);

var _AsyncStrategy = require('../../enums/AsyncStrategy');

var _AsyncStrategy2 = _interopRequireDefault(_AsyncStrategy);

var _DebounceStatus = require('../../enums/DebounceStatus');

var _DebounceStatus2 = _interopRequireDefault(_DebounceStatus);

var _utils = require('./utils');

var _ids = require('../ids');

var _utils2 = require('../utils');

var _buildErrorMessage = require('../buildErrorMessage');

var _buildErrorMessage2 = _interopRequireDefault(_buildErrorMessage);

var _strategies = require('./strategies');

var _strategies2 = _interopRequireDefault(_strategies);

var _asyncStrategies = require('./async-strategies');

var _asyncStrategies2 = _interopRequireDefault(_asyncStrategies);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

/**
 * @desc Performs synchronous single field validation.
 *       It includes calculation of the sync strategy for the field.
 *
 */
function performSyncFieldValidation(field, value, data, event, stateContainer) {
  var ON_FIRST_BLUR = _Strategy2.default.ON_FIRST_BLUR,
      ON_FIRST_SUCCESS_OR_FIRST_BLUR = _Strategy2.default.ON_FIRST_SUCCESS_OR_FIRST_BLUR,
      DEFAULT = _Strategy2.default.DEFAULT;

  // Figuring out defined strategy

  var strategy = field.strategy || stateContainer.getPropsLevelStrategy() || DEFAULT;

  // Figuring out defined async strategy
  var asyncStrategy = field.asyncStrategy || stateContainer.getPropsLevelAsyncStrategy() || _AsyncStrategy2.default.DEFAULT;

  // Ignoring blur event, if strategies don't care about blur
  var blurConcernedStrategy = [ON_FIRST_BLUR, ON_FIRST_SUCCESS_OR_FIRST_BLUR].includes(strategy);
  var blurConcernedAsyncStrategy = !!field.validateAsync && asyncStrategy === _AsyncStrategy2.default.ON_BLUR;

  if (event && (0, _utils2.isBlurEvent)(event) && !blurConcernedStrategy && !blurConcernedAsyncStrategy) {
    return null;
  }

  // Also ignoring blur event, if:
  //   - field doesn't have async validation w/ ON_BLUR strategy
  //   - and sync strategy is concerned about the first blur,
  //     but it's not the first blur event for this field
  if (event && (0, _utils2.isBlurEvent)(event) && blurConcernedStrategy && !blurConcernedAsyncStrategy && (stateContainer.getBluredField(field.id) || stateContainer.getEmittedField(field.id))) {
    return null;
  }

  // If form was subitted, then switching to ON_FIRST_SUBMIT strategy
  // Otherwise using provided strategy
  var validateWithStrategy = stateContainer.getFormWasSubmitted() ? _strategies2.default.onFirstSubmit : _strategies2.default[strategy];

  if (!validateWithStrategy) {
    throw new Error((0, _buildErrorMessage2.default)({
      layerId: stateContainer.getLayerId(),
      fieldId: field.id,
      message: 'Unknown strategy: ' + strategy + '. Most likely it\'s a spelling issue.'
    }));
  }

  // Validating with the calculated strategy
  var validatedWithStrategy = validateWithStrategy(field, value, data, event, stateContainer);

  // Before emit one more case must be handled:
  // if this is `change` event and validation was successful,
  // but field has also async validation w/ ON_BLUR stategy.
  // In this case not emitting anything until the blur.
  return blurConcernedAsyncStrategy && validatedWithStrategy.valid && event && (0, _utils2.isChangeEvent)(event) ? (0, _utils.buildEmptyValidationResults)() : validatedWithStrategy;
}

/**
 * @desc Performs asynchronous single field validation.
 *       It includes calculation of the async strategy for the field.
 *
 */
function performAsyncFieldValidation(syncValidationResults, field, value, event, stateContainer) {
  // Don't perform async validation if:
  //   - there is no async validator
  //   - value is empty (no reason to validate empty value on remote)
  //   - value is invalid localy (no reason to validate invalid value on remote)
  //   - sync validation said `don't-provide-feedback-yet`
  var noAsyncValidator = !field.validateAsync;
  var noValue = !value && value !== 0;
  var isSyncValidationInvalid = syncValidationResults && !syncValidationResults.valid;
  var noEmittableSyncResults = syncValidationResults && syncValidationResults.valid === null && syncValidationResults.status === null && syncValidationResults.message === null;

  if (noAsyncValidator || noValue || isSyncValidationInvalid || noEmittableSyncResults) {
    return syncValidationResults;
  }

  // Figuring out async strategy
  var asyncStrategy = field.asyncStrategy || stateContainer.getPropsLevelAsyncStrategy() || _AsyncStrategy2.default.DEFAULT;

  // If it's ON_BLUR and event.type is blur -> validate
  if (event && (0, _utils2.isBlurEvent)(event) && asyncStrategy === _AsyncStrategy2.default.ON_BLUR) {
    return _asyncStrategies2.default.onBlur(field, value, stateContainer);
  }

  // Don't perform async validation if:
  //   - validation is skipped by sync validator (null value returned)
  //   - it's a change event, but async strategy is ON_BLUR
  var isSyncValidationSkipped = syncValidationResults === null;
  var isOnChangeEventButOnBlurStrategy = event && (0, _utils2.isChangeEvent)(event) && asyncStrategy === _AsyncStrategy2.default.ON_BLUR;

  if (isSyncValidationSkipped || isOnChangeEventButOnBlurStrategy) {
    return syncValidationResults;
  }

  // Otherwise -> debounced ON_CHANGE
  return _asyncStrategies2.default.onChange(field, value, stateContainer);
}

/**
 * @desc Controller of sync & async validations.
 *
 */
function performFieldValidation(parentField, parentValue, parentEvent, stateContainer) {
  if (parentField.linkedFields && !Array.isArray(parentField.linkedFields)) {
    throw new Error((0, _buildErrorMessage2.default)({
      layerId: stateContainer.getLayerId(),
      fieldId: parentField.id,
      message: '`linkedFields` must be an Array.'
    }));
  }

  // Figuring out set of fields to validate:
  //   - parent field (which's changed triggered validation)
  //   - linked fields (if any)
  //
  // NOTE: Theoretically linked field might has its own linked fields.
  //       Something to keep in mind, but ignoring until first use case.
  var fields = parentField.linkedFields ? parentField.linkedFields.reduce(function (allFields, keyPath) {
    var fieldId = (0, _ids.buildFieldIdFromUserKeyPath)(keyPath);
    return allFields.concat(stateContainer.getNormalizedField(fieldId));
  }, [parentField]) : [parentField];

  // As we (might) have multiple fields to validate,
  // we don't want to trigger state update for each field separately,
  // but we want to batch updates and set next state within single transaction:
  // set sync valiudation results + start processing async ones.
  // Later on, when async results will be resolved ->
  // state will be updated again for each resolved field separetly.
  // Sync state will be packed under `nextSyncState` key,
  // Promises w/ ongoing async validations - under `ongoingAsyncValidations` key.
  var origin = { nextSyncState: {}, ongoingAsyncValidations: {} };

  return fields.reduce(function (results, field) {
    var _extends5;

    var isParentField = field.id === parentField.id;

    var currentData = stateContainer.getData();

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var value = isParentField ? parentValue : (0, _utils2.fetchProp)(currentData, field.keyPath);
    var event = isParentField ? parentEvent : null;

    // If it's a linked field, data must be updated w/ the next value of parent field.
    // Not doing this for parent field as it affects perf.
    var data = isParentField ? currentData : _.set(_.cloneDeep(currentData), parentField.keyPath, parentValue);

    // Performing sync -> async validations
    var syncValidationResults = performSyncFieldValidation(field, value, data, event, stateContainer);
    var finalValidationResults = performAsyncFieldValidation(syncValidationResults, field, value, event, stateContainer);

    // If no results returned or it's a debounced results -> ignoring them
    if (!finalValidationResults || finalValidationResults.status === _DebounceStatus2.default.DEBOUNCED) {
      return results;
    }

    var stateId = (0, _ids.buildFieldValidationStateId)(field.id);

    // If it's a Promise, it means that we should set state of this field to processing
    // and store Promise to update validation state later on (in the core component)
    if ((0, _utils2.isPromise)(finalValidationResults)) {
      var _extends2, _extends3;

      return {
        nextSyncState: _extends({}, results.nextSyncState, (_extends2 = {}, _extends2[stateId] = (0, _utils.buildIntermediateAsyncValidationResults)(), _extends2)),
        ongoingAsyncValidations: _extends({}, results.ongoingAsyncValidations, (_extends3 = {}, _extends3[stateId] = finalValidationResults, _extends3))
      };
    }

    // If it's a message from the debouncer w/ the status INITIALIZED ->
    // just starting processing the field.
    // Debouncer will notify state container about triggered request directly
    // and pass Promise within notification
    if (finalValidationResults.status === _DebounceStatus2.default.INITIALIZED) {
      var _extends4;

      return {
        nextSyncState: _extends({}, results.nextSyncState, (_extends4 = {}, _extends4[stateId] = (0, _utils.buildIntermediateAsyncValidationResults)(), _extends4)),
        ongoingAsyncValidations: results.ongoingAsyncValidations
      };
    }

    // Otherwise it's a results of sync validation. Just merging them.
    return {
      nextSyncState: _extends({}, results.nextSyncState, (_extends5 = {}, _extends5[stateId] = finalValidationResults, _extends5)),
      ongoingAsyncValidations: results.ongoingAsyncValidations
    };
  }, origin);
}

/**
 * @desc Performs fields validation on mount.
 *
 */
function performOnMountValidation(fields, stateContainer) {
  var layerId = stateContainer.getLayerId();
  var data = stateContainer.getData();
  var normalizedFields = stateContainer.getNormalizedFields();
  var propsLevelStatuses = stateContainer.getPropsLevelStatuses();
  var propsLevelStrategy = stateContainer.getPropsLevelStrategy();
  var propsLevelAsyncStrategy = stateContainer.getPropsLevelAsyncStrategy();

  var origin = { nextSyncState: {}, ongoingAsyncValidations: {} };

  return normalizedFields.filter(function (field) {
    return (field.strategy || propsLevelStrategy) === _Strategy2.default.ON_MOUNT;
  }).reduce(function (_ref, field) {
    var _extends7, _extends8;

    var nextSyncState = _ref.nextSyncState,
        ongoingAsyncValidations = _ref.ongoingAsyncValidations;

    stateContainer.setEmittedField(field.id);

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var value = (0, _utils2.fetchProp)(data, field.keyPath);
    var stateId = (0, _ids.buildFieldValidationStateId)(field.id);

    var fieldNextSyncValidationState = (0, _utils.buildCompleteSyncValidationResults)(field, value, data, propsLevelStatuses, layerId);

    // Don't perform async validation if:
    //   - there is no async validator
    //   - value is empty (no reason to validate empty value on remote)
    //   - value is invalid localy (no reason to validate invalid value on remote)
    var isAsync = !!field.validateAsync && (!!value || value === 0) && fieldNextSyncValidationState.valid;

    if (!isAsync) {
      var _extends6;

      return {
        nextSyncState: _extends({}, nextSyncState, (_extends6 = {}, _extends6[stateId] = fieldNextSyncValidationState, _extends6)),
        ongoingAsyncValidations: ongoingAsyncValidations
      };
    }

    var asyncStrategy = field.asyncStrategy || propsLevelAsyncStrategy || _AsyncStrategy2.default.DEFAULT;

    // If asyncStrategy is ON_CHANGE,
    // then validator on normalized field is debounced.
    // We'll use original one.
    var asyncValidator = asyncStrategy === _AsyncStrategy2.default.ON_CHANGE ? (0, _utils2.fetchProp)(fields, field.keyPath).validateAsync : field.validateAsync;

    return {
      nextSyncState: _extends({}, nextSyncState, (_extends7 = {}, _extends7[stateId] = (0, _utils.buildIntermediateAsyncValidationResults)(), _extends7)),
      ongoingAsyncValidations: _extends({}, ongoingAsyncValidations, (_extends8 = {}, _extends8[stateId] = asyncValidator(value).then(function (results) {
        return {
          resolution: (0, _utils.buildCompleteAsyncValidationResults)(field.id, results, propsLevelStatuses, layerId),
          forValue: value
        };
      }), _extends8))
    };
  }, origin);
}

/**
 * @desc Performs fields validation on form submission.
 *       We don't want to perform async validations here
 *       as those will be performed on the server anyway - all within single request.
 *       However we should respect the current state of async validations.
 *       That being said, algoritm looks like this:
 *         - perform batch synchronous validation of all fields
 *         - resolve merge conflicts w/ existing state using the following regulations:
 *           - no current state -> just merge-in next one
 *           - current failure & next failure -> take next one
 *           - current success & next failure -> take next one
 *           - current failure & next success -> current isAsync ? keep current : take next one
 *           - current success & next success -> current isAsync ? keep current : take next one
 *         - return next composite state + validity resolution
 *
 */
function performOnSubmitValidation(data, state, fields, statuses, layerId) {
  var origin = { validationState: {}, isValid: true };

  return fields.reduce(function (_ref2, field) {
    var _extends9;

    var validationState = _ref2.validationState,
        isValid = _ref2.isValid;

    // $FlowIgnoreMe: We're making sure that value at keyPath is not an object on normalization
    var value = (0, _utils2.fetchProp)(data, field.keyPath);
    var fieldValidationStateId = (0, _ids.buildFieldValidationStateId)(field.id);

    var fieldCurrentValidationState = state[fieldValidationStateId];
    var fieldNextSyncValidationState = (0, _utils.buildCompleteSyncValidationResults)(field, value, data, statuses, layerId);

    var fieldNextValidationState = fieldCurrentValidationState && fieldNextSyncValidationState.valid && fieldCurrentValidationState.isAsync ? fieldCurrentValidationState : fieldNextSyncValidationState;

    return {
      isValid: !isValid ? isValid : !!fieldNextValidationState.valid,
      validationState: _extends({}, validationState, (_extends9 = {}, _extends9[fieldValidationStateId] = fieldNextValidationState, _extends9))
    };
  }, origin);
}