'use strict';

exports.__esModule = true;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.normalizeValidationResults = normalizeValidationResults;
exports.applySyncValidation = applySyncValidation;
exports.getDefaultStatuses = getDefaultStatuses;
exports.buildCompleteSyncValidationResults = buildCompleteSyncValidationResults;
exports.buildCompleteAsyncValidationResults = buildCompleteAsyncValidationResults;
exports.buildIntermediateAsyncValidationResults = buildIntermediateAsyncValidationResults;
exports.buildEmptyValidationResults = buildEmptyValidationResults;
exports.debounce = debounce;

var _DefaultStatus = require('../../../enums/DefaultStatus');

var _DefaultStatus2 = _interopRequireDefault(_DefaultStatus);

var _DebounceStatus = require('../../../enums/DebounceStatus');

var _DebounceStatus2 = _interopRequireDefault(_DebounceStatus);

var _buildErrorMessage = require('../../buildErrorMessage');

var _buildErrorMessage2 = _interopRequireDefault(_buildErrorMessage);

var _utils = require('../../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @desc It is possible to return validation result as simple boolean,
 *       as well as Object w/ additional data (i.e. `message` & `status`).
 *       This function normalizes validation results to Object shape.
 *
 */
function normalizeValidationResults(results, layerId, fieldId) {
  if (results === true) {
    return { valid: true };
  } else if (results === false) {
    return { valid: false };
  }

  if (!(0, _utils.isDefined)(results.valid)) {
    throw new Error((0, _buildErrorMessage2.default)({
      layerId: layerId,
      fieldId: fieldId,
      message: ["Can't find `valid` key in the results object from validator.", "Make sure it's defined and its value is Boolean"]
    }));
  }

  var normalizedResults = {
    valid: results.valid
  };

  if (results.status) normalizedResults.status = results.status;
  if (results.message) normalizedResults.message = results.message;

  return normalizedResults;
}

/**
 * @desc Applies sync validator and returns normalized validation results.
 *
 */
function applySyncValidation(validator, value, data, layerId, fieldId) {
  if (!validator) return { valid: true };

  if (validator && !(0, _utils.isFunction)(validator)) {
    throw new Error((0, _buildErrorMessage2.default)({
      layerId: layerId,
      fieldId: fieldId,
      message: 'Validator must be a function.'
    }));
  }

  var results = validator(value, data);

  return normalizeValidationResults(results, layerId, fieldId);
}

/**
 * @desc Calculates default success & failure statuses.
 *
 */
// $FlowIssue: https://github.com/facebook/flow/issues/2977
function getDefaultStatuses() {
  var statuses = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  return {
    success: statuses.success || _DefaultStatus2.default.SUCCESS,
    failure: statuses.failure || _DefaultStatus2.default.FAILURE
  };
}

/**
 * @desc Builds Object w/ validation results from synchronous validation.
 *       This results will be emitted to the view component.
 *
 */
function buildCompleteSyncValidationResults(field, value, data, statuses, layerId) {
  var defaultStatuses = getDefaultStatuses(statuses);
  var validationState = applySyncValidation(field.validate, value, data, layerId, field.id);

  if (!validationState.valid) {
    validationState.status = validationState.status || defaultStatuses.failure;
  } else if (value) {
    validationState.status = validationState.status || defaultStatuses.success;
  }

  return validationState;
}

/**
 * @desc Builds Object w/ validation results from asynchronous validation.
 *       This results will be emitted to the view component.
 *
 */
function buildCompleteAsyncValidationResults(fieldId, results, statuses, layerId) {
  var defaultStatuses = getDefaultStatuses(statuses);
  var validationState = _extends({}, normalizeValidationResults(results, layerId, fieldId), {
    isAsync: true
  });

  if (!validationState.valid) {
    validationState.status = validationState.status || defaultStatuses.failure;
  } else {
    validationState.status = validationState.status || defaultStatuses.success;
  }

  // $FlowIssue: exact type + destructuring: https://github.com/facebook/flow/issues/2405
  return validationState;
}

/**
 * @desc Builds Object w/ intermediate async validation results.
 *       These results will be emitted to the view component,
 *       when either debounced or non-debounced async validator
 *       is triggered and we're waiting for remote validation to be resolved.
 *
 */
function buildIntermediateAsyncValidationResults() {
  return {
    valid: null,
    status: null,
    message: null,
    isProcessing: true
  };
}

/**
 * @desc Builds Object w/ empty validation results.
 *       These results will be emitted to the view component,
 *       but it means that Validation Layer doesn't emit
 *       any results yet according to strategy.
 *
 */
function buildEmptyValidationResults() {
  return {
    valid: null,
    status: null,
    message: null
  };
}

/**
 * @desc Debounce async validation calls.
 *       Based on lodash.debounce.
 *
 */

/* eslint-disable no-use-before-define, consistent-return */

function debounce(async) {
  var wait = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

  var lastContext = void 0;
  var lastValue = void 0;
  var lastFieldId = void 0;
  var lastStateContainer = void 0;
  var lastCallTime = void 0;
  var timerId = void 0;

  function shouldInvoke(time) {
    var timeSinceLastCall = time - lastCallTime;

    return lastCallTime === undefined || timeSinceLastCall >= wait || timeSinceLastCall < 0;
  }

  function initAsync() {
    // Start the timer for the invoke async
    timerId = setTimeout(timerExpired, wait);
    // Report initialized
    return { status: _DebounceStatus2.default.INITIALIZED };
  }

  function remainingWait(time) {
    var timeSinceLastCall = time - lastCallTime;
    return wait - timeSinceLastCall;
  }

  function timerExpired() {
    var time = Date.now();

    if (shouldInvoke(time)) {
      return invokeAsync(time);
    }

    timerId = setTimeout(timerExpired, remainingWait(time));
  }

  function invokeAsync() {
    timerId = undefined;

    if (lastValue) {
      return triggerRequest();
    }
  }

  function triggerRequest() {
    var context = lastContext;
    var fieldId = lastFieldId;
    var value = lastValue;
    var stateContainer = lastStateContainer;

    lastContext = lastFieldId = lastValue = lastStateContainer = undefined;

    if (!fieldId || !stateContainer) throw new Error();

    // Notify layer about triggered request
    return stateContainer.notifyAsync(fieldId, value, async.call(context, value));
  }

  return function debounced(fieldId, value, stateContainer) {
    var time = Date.now();
    var isInvoking = shouldInvoke(time);

    lastCallTime = time;
    lastContext = this;

    lastFieldId = fieldId;
    lastValue = value;
    lastStateContainer = stateContainer;

    // Initialize if it's a first ivokation
    if (isInvoking && timerId === undefined) {
      return initAsync(lastCallTime);
    }

    // Start timer if it's not done yet
    if (timerId === undefined) {
      timerId = setTimeout(timerExpired, wait);
    }

    // Report debounced
    return { status: _DebounceStatus2.default.DEBOUNCED };
  };
}